import pandas as pd
import numpy as np
from datetime import datetime
import pandas_datareader as pdr


class CovidResilience:
    def __init__(self, stock_index = "STOXX 600", start_date = "2020-01-19", end_date = "2021-01-19", tickers = None):
        self.stock_index = stock_index
        self.start_date = start_date
        self.end_date = end_date
        self.data = None
        self.informations = None
        self.stock_index_tickers = {"CAC 40":"^FCHI","BEL 20":"^BFX", "S&P 500":"^GSPC", "STOXX 600":"^STOXX"}
        self.returns = None #on the period
        self.daily_returns = None #daily
        self.path_to_data = None
        self.stocks_selection = "covid_resilient"
        self.relevant_data = None
        self.relevant_data_returns = None
        self.volatilities = None
        self.rslt = None
        if tickers != None:
            self.tickers = tickers
        else :
            self.tickers = ['NOVO-B.CO',
                            'EQNR.OL',
                            'ATCO-A.ST',
                            'ORSTED.CO',
                            'INVE-B.ST',
                            'VOLV-B.ST',
                            'ERIC-B.ST',
                            'HM-B.ST',
                            'MAERSK-B.CO',
                            'NESN.SW',
                            'VWS.CO',
                            'NDA-SE.ST',
                            'HEXA-B.ST',
                            'ROG.SW',
                            'SAND.ST',
                            'DNB.OL',
                            'MC.PA',
                            'ASSA-B.ST',
                            'DSV.CO',
                            'NOVN.SW',
                            'EQT.ST',
                            'TEL.OL',
                            'SEB-A.ST',
                            'COLO-B.CO',
                            'EPI-A.ST',
                            'ESSITY-B.ST',
                            'ASML.AS',
                            'EVO.ST',
                            'GMAB.CO',
                            'SWED-A.ST',
                            'SHB-A.ST',
                            'OR.PA',
                            'PRX.AS',
                            'TELIA.ST',
                            'NIBE-B.ST',
                            'CARL-B.CO',
                            'SAP.DE',
                            'RDSA.AS',
                            'LATO-B.ST',
                            'INDU-A.ST',
                            'BHP.L',
                            'KINV-B.ST',
                            'ULVR.L',
                            'ABI.BR',
                            'LIN.DE',
                            'LUND-B.ST',
                            'SKF-B.ST',
                            'YAR.OL',
                            'SAN.PA',
                            'NZYM-B.CO',
                            'SIE.DE',
                            'SCA-B.ST',
                            'AZN.L',
                            'SWMA.ST',
                            'RIO.L',
                            'FP.PA',
                            'GJF.OL',
                            'MOWI.OL',
                            'ALFA.ST',
                            'DANSKE.CO',
                            'SKA-B.ST',
                            'RMS.PA',
                            'ADE.OL',
                            'ENEL.MI',
                            'AKER.OL',
                            'ORK.OL',
                            'HSBA.L',
                            'ALV.DE',
                            'NHY.OL',
                            'IC-A.ST',
                            'BALD-B.ST',
                            'ITX.MC',
                            'VOW3.DE',
                            'TEL2-B.ST',
                            'BOL.ST',
                            'IBE.MC',
                            'LIFCO-B.ST',
                            'SINCH.ST',
                            'CHR.CO',
                            'SCHA.OL',
                            'DTE.DE',
                            'AIR.PA',
                            'LUNE.ST',
                            'GSK.L',
                            'SU.PA',
                            'KER.PA',
                            'GN.CO',
                            'DGE.L',
                            'AMBU-B.CO',
                            'HOLM-B.ST',
                            'MRK.DE',
                            'INDT.ST',
                            'HUSQ-B.ST',
                            'BATS.L',
                            'AI.PA',
                            'ELUX-B.ST',
                            'BP.L',
                            'PNDORA.CO',
                            'BAS.DE',
                            'TOM.OL',
                            'TRYG.CO',
                            'DAI.DE',
                            'ADS.DE',
                            'DEMANT.CO',
                            'SAGA-B.ST',
                            'CAST.ST',
                            'SCATC.OL',
                            'ABBN.SW',
                            'ZURN.SW',
                            'SALM.OL',
                            'BNP.PA',
                            'SWEC-B.ST',
                            'TREL-B.ST',
                            'ADYEN.AS',
                            'EL.PA',
                            'GETI-B.ST',
                            'BAYN.DE',
                            'ROCK-B.CO',
                            'HEIA.AS',
                            'SAF.PA',
                            'UBSG.SW',
                            'DG.PA',
                            'SOBI.ST',
                            'SECU-B.ST',
                            'CFR.SW',
                            'BEIJ-B.ST',
                            'CS.PA',
                            'SHL.DE',
                            'NESTE.HE',
                            'SAN.MC',
                            'NEL.OL',
                            'IFX.DE',
                            'EKTA-B.ST',
                            'RB.L',
                            'BMW.DE',
                            'FABG.ST',
                            'LONN.SW',
                            'SBB-B.ST',
                            'DSY.PA',
                            'WALL-B.ST',
                            'AXFO.ST',
                            'PHIA.AS',
                            'RI.PA',
                            'AAK.ST',
                            'BN.PA',
                            'EDF.PA',
                            'PRU.L',
                            'ISP.MI',
                            'BAKKA.OL',
                            'GLEN.L',
                            'HEN3.DE',
                            'REL.L',
                            'ENTRA.OL',
                            'SIM.CO',
                            'SIKA.SW',
                            'KNEBV.HE',
                            'VOD.L',
                            'RACE.MI',
                            'THULE.ST',
                            'DOM.ST',
                            'MUV2.DE',
                            'ENGI.PA',
                            'AAL.L',
                            'GIVN.SW',
                            'STB.OL',
                            'LSEG.L',
                            'ENI.MI',
                            'SAAB-B.ST',
                            'ADDT-B.ST',
                            'ALC.SW',
                            'VNA.DE',
                            'NENT-B.ST',
                            'BILL.ST',
                            'LHN.SW',
                            'NG.L',
                            'RBREW.CO',
                            'VIV.PA',
                            'HPOL-B.ST',
                            'NETC.CO',
                            'INGA.AS',
                            'STM.MI',
                            'CSGN.SW',
                            'ACA.PA',
                            'DIM.PA',
                            'CRG.IR',
                            'WIHL.ST',
                            'FLTR.IR',
                            'PGHN.SW',
                            'AF-B.ST',
                            'DHER.DE',
                            'VER.VI',
                            'BBVA.MC',
                            'SRT3.DE',
                            'SREN.SW',
                            'AD.AS',
                            'TOP.CO',
                            'DSM.AS',
                            'DB1.DE',
                            'ORA.PA',
                            'LLOY.L',
                            'BARC.L',
                            'SCHP.SW',
                            'RWE.DE',
                            'AMS.MC',
                            'CPG.L',
                            'KBC.BR',
                            'EXPN.L',
                            'SCMN.SW',
                            'KNIN.SW',
                            'ZAL.DE',
                            'SGRE.MC',
                            'ELE.MC',
                            'TSCO.L',
                            'CON.DE',
                            'BEI.DE',
                            'EOAN.DE',
                            'ENR.DE',
                            'CLNX.MC',
                            'G.MI',
                            'SGO.PA',
                            'HEIO.AS',
                            'FRE.DE',
                            'LR.PA',
                            'MT.AS',
                            'CAP.PA',
                            'EDP.LS',
                            'TEF.MC',
                            'GEBN.SW',
                            'WLN.PA',
                            'EDPR.LS',
                            'SGSN.SW',
                            'AENA.MC',
                            'FERG.L',
                            'FME.DE',
                            'LISN.SW',
                            'KYGA.L',
                            'STLA.MI',
                            'SAMPO.HE',
                            'NTGY.MC',
                            'OCDO.L',
                            'ISS.CO',
                            'ML.PA',
                            'EMSN.SW',
                            'DBK.DE',
                            'NWG.L',
                            'FORTUM.HE',
                            'WKL.AS',
                            'NOKIA.HE',
                            'KBX.DE',
                            'ABF.L',
                            'RYA.L',
                            'PAH3.DE',
                            'UCG.MI',
                            'UCB.BR',
                            'TEP.PA',
                            'AKZA.AS',
                            'HNR1.DE',
                            'AHT.L',
                            'UPM.HE',
                            'STMN.SW',
                            'CCL.L',
                            'JDEP.AS',
                            'HO.PA',
                            'LGEN.L',
                            'EXO.MI',
                            'SSE.L',
                            'BA.L',
                            'FER.MC',
                            'GRF.MC',
                            'IMB.L',
                            'DWNI.DE',
                            'LOGN.SW',
                            'STAN.L',
                            'CNHI.MI',
                            'ANTO.L',
                            'SRG.MI',
                            'ERF.PA',
                            'GLE.PA',
                            'SOON.SW',
                            'ALO.PA',
                            'SN.L',
                            'BIM.PA',
                            'BTA.L',
                            'SLHN.SW',
                            'REP.MC',
                            'AMUN.PA',
                            'SY1.DE',
                            'AV.L',
                            'GBLB.BR',
                            'VIE.PA',
                            'EN.PA',
                            'TKWY.AS',
                            'CABK.MC',
                            'UHR.SW',
                            'PUM.DE',
                            'HEI.DE',
                            'STERV.HE',
                            'EVK.DE',
                            'BAER.SW',
                            'MONC.MI',
                            'CA.PA',
                            'OMV.VI',
                            'TRN.MI',
                            'NN.AS',
                            'KSP.IR',
                            'SW.PA',
                            'SGRO.L',
                            'EBS.VI',
                            'EDEN.PA',
                            'AFX.DE',
                            'MTX.DE',
                            'ATL.MI',
                            'UMI.BR',
                            'HFG.DE',
                            'ARGX.BR',
                            'BARN.SW',
                            'III.L',
                            'KPN.AS',
                            'SEV.PA',
                            'PST.MI',
                            'AVV.L',
                            'UN01.DE',
                            'ASM.AS',
                            'PUB.PA',
                            '1COV.DE',
                            'NXT.L',
                            'SOLB.BR',
                            'BNR.DE',
                            'UBI.PA',
                            'SKG.L',
                            'BOL.PA',
                            'BVI.PA',
                            'CPR.MI',
                            'QIA.DE',
                            'RNO.PA',
                            'RAND.AS',
                            'KN.PA',
                            'KGX.DE',
                            'WPP.L',
                            'RTO.L',
                            'ADEN.SW',
                            'DIA.MI',
                            'NEXI.MI',
                            'HLMA.L',
                            'ADP.PA',
                            'CNP.PA',
                            'ILD.PA',
                            'GFC.PA',
                            'SOF.BR',
                            'JMT.LS',
                            'AT1.DE',
                            'REC.MI',
                            'SDR.L',
                            'ACS.MC',
                            'VIFN.SW',
                            'MRO.L',
                            'RAA.DE',
                            'IHG.L',
                            'ITRK.L',
                            'INW.MI',
                            'LEG.DE',
                            'CRDA.L',
                            'RR.L',
                            'AGS.BR',
                            'MNDI.L',
                            'PSN.L',
                            'ADM.L',
                            'REE.MC',
                            'TMV.DE',
                            'TEMN.SW',
                            'KESKOB.HE',
                            'CCH.L',
                            'SPX.L',
                            'URW.AS',
                            'FBK.MI',
                            'BNZL.L',
                            'ABN.AS',
                            'JD.L',
                            'TEN.MI',
                            'BCVN.SW',
                            'AC.PA',
                            'FRES.L',
                            'IAG.L',
                            'FGR.PA',
                            'TIT.MI',
                            'INF.L',
                            'SK.PA',
                            'AMP.MI',
                            'ELISA.HE',
                            'POLY.L',
                            'BALN.SW',
                            'GALP.LS',
                            'PRY.MI',
                            'HL.L',
                            'AM.PA',
                            'AGN.AS',
                            'VACN.SW',
                            'GET.PA',
                            'ORP.PA',
                            'BC8.DE',
                            'EVR.L',
                            'AKE.PA',
                            'FR.PA',
                            'ATO.PA',
                            'ENT.L',
                            'THG.L',
                            'FDJ.PA',
                            'CBK.DE',
                            'RCO.PA',
                            'MOCORP.HE',
                            'BDEV.L',
                            'ANA.MC',
                            'BRBY.L',
                            'RSA.L',
                            'UTDI.DE',
                            'ELI.BR',
                            'NEM.DE',
                            'ATCO-A.ST',
                            'PHNX.L',
                            'G24.DE',
                            'O2D.DE',
                            'COLR.BR',
                            'COV.PA',
                            'SIGN.SW',
                            'SPSN.SW',
                            'MB.MI',
                            'SLA.L',
                            'ENX.PA',
                            'CLN.SW',
                            'UU.L',
                            'WTB.L',
                            'STJ.L',
                            'LHA.DE',
                            'IPN.PA',
                            'SGE.L',
                            'TKA.DE',
                            'AMS.SW',
                            'IMCD.AS',
                            'FPE3.DE',
                            'SMIN.L',
                            'TW.L',
                            'SOI.PA',
                            'EO.PA',
                            'RBI.VI',
                            'HIK.L',
                            'GLPG.AS',
                            'PROX.BR',
                            'DCC.L',
                            'LXS.DE',
                            'VOE.VI',
                            'SVT.L',
                            'AVST.L',
                            'BKG.L',
                            'KGF.L',
                            'ORNBV.HE',
                            'LI.PA',
                            'AUTO.L',
                            'G1A.DE',
                            'WRT1V.HE',
                            'RMV.L',
                            'SBRY.L',
                            'JMAT.L',
                            'VPK.AS',
                            'TECN.SW',
                            'PSPN.SW',
                            'SMDS.L',
                            'SCR.PA',
                            'LIGHT.AS',
                            'WDP.BR',
                            'PSON.L',
                            'WEIR.L',
                            'HELN.SW',
                            'BME.L',
                            'MNG.L',
                            'ICP.L',
                            'LAND.L',
                            'VAR1.DE',
                            'ENG.MC',
                            'ASRNL.AS',
                            'EVD.DE',
                            'IDIA.SW',
                            'EVT.DE',
                            'FI.SW',
                            'FHZN.SW',
                            'BEAN.SW',
                            'RF.PA',
                            'HER.MI',
                            'SWON.SW',
                            'HUH1V.HE',
                            'ICAD.PA',
                            'MF.PA',
                            'KOJAMO.HE',
                            'DLG.L',
                            'FTI.PA',
                            'IP.MI',
                            'ITV.L',
                            'ACKB.BR',
                            'MRW.L',
                            'SRAIL.SW',
                            'BKT.MC',
                            'ECM.L',
                            'DUFN.SW',
                            'BESI.AS',
                            'WAF.DE',
                            'TYRES.HE',
                            'A2A.MI',
                            'BLND.L',
                            'HWDN.L',
                            'COL.MC',
                            'RUI.PA',
                            'AALB.AS',
                            'RXL.PA',
                            'ROSE.SW',
                            'ANDR.VI',
                            'IG.MI',
                            'PNN.L',
                            'GFS.L',
                            'CTEC.L',
                            'DPH.L',
                            'RMG.L',
                            'TUI.L',
                            'RHM.DE',
                            'VALMT.HE',
                            'UTG.L',
                            'DLG.DE',
                            'TEG.DE',
                            'MRL.MC',
                            'BIRG.IR',
                            'HSV.L',
                            'DLN.L',
                            'BG.VI',
                            'GYC.DE',
                            'BWY.L',
                            'SXS.L',
                            'GAW.L',
                            'TPK.L',
                            'HSX.L',
                            'SESG.PA',
                            'COFB.BR',
                            'IMI.L',
                            'LDO.MI',
                            'IWG.L',
                            'MGGT.L',
                            'WIE.VI',
                            'IGG.L',
                            'ATE.PA',
                            'ALLN.SW',
                            'AED.BR',
                            'MOR.DE',
                            'ASHM.L',
                            'BBOX.L',
                            'PSM.DE',
                            'TATE.L',
                            'CMBN.SW',
                            'ADJ.DE',
                            'GLB.L',
                            'GALE.SW',
                            'GNS.L',
                            'SPIE.PA',
                            'CNA.L',
                            'ROR.L',
                            'ELIS.PA',
                            'SBMO.AS',
                            'BAMI.MI',
                            'WMH.L',
                            'GXI.DE',
                            'VIS.MC',
                            'SOP.PA',
                            'QLT.L',
                            'SFZN.SW',
                            'CRBN.AS',
                            'DPLM.L',
                            'MKS.L',
                            'AOX.DE',
                            'INCH.L',
                            'HAS.L',
                            'FNTN.DE',
                            'CSP.L',
                            'EMG.L',
                            'SAB.MC',
                            'BEZ.L',
                            'CBG.L',
                            'VCT.L',
                            'TRN.L',
                            'LMP.L',
                            'BVIC.L',
                            'AGR.L',
                            'PHP.L',
                            'UDG.L',
                            'GRI.L',
                            'CWK.L']

    def load_data(self):
        if self.path_to_data:
            self.data = pd.read_csv(self.path_to_data, parse_dates = True, index_col=0)
        else:
            stocks_ticker = [self.stock_index_tickers[self.stock_index]] + self.tickers
            for ticker in stocks_ticker :
                if ticker == stocks_ticker[0]:
                    self.data = pdr.get_data_yahoo(ticker, start=datetime.strptime(self.start_date, '%Y-%m-%d'), end=datetime.strptime(self.end_date, '%Y-%m-%d'))[["Close"]].fillna(method="bfill")
                    self.data.columns = [ticker]
                else:
                    try:
                        try:
                            self.data[ticker] = pdr.get_data_yahoo(ticker, start=datetime.strptime(self.start_date, '%Y-%m-%d'), end=datetime.strptime(self.end_date, '%Y-%m-%d')).Close.fillna(method="bfill")
                        except:
                            self.data[ticker] = pdr.get_data_yahoo(ticker.replace("-A","A"), start=datetime.strptime(self.start_date, '%Y-%m-%d'), end=datetime.strptime(self.end_date, '%Y-%m-%d')).Close.fillna(method="bfill")
                    except:
                        print(ticker + " Not found")
          #self.data.to_csv("data.csv")

    def stocks_selector(self):
      if self.stocks_selection == "covid_resilient":
          relevant_dates = ["2020-01-20", "2020-03-18","2020-10-28","2021-01-19"]
          self.relevant_data = self.data.loc[relevant_dates, :]

    def compute_returns(self):
        # Compute daily returns on the whole period and at the same time
        self.daily_returns = pd.DataFrame((self.data.iloc[:,0]-self.data.iloc[:,0].shift(1))/self.data.iloc[:,0].shift(1))
        for col in self.data.columns[1:]:
            self.daily_returns[col] = (self.data[col]-self.data[col].shift(1))/self.data[col].shift(1)
        self.daily_returns.fillna(0, inplace=True)
        #daily_returns.to_excel("daily_returns"+self.stock_index+".xls")
        #self.daily_returns.to_csv("returns.csv")
        # Compute returns of relevant data in case we use the covid_resilient selector
        if self.stocks_selection == "covid_resilient":
            self.relevant_data_returns = pd.DataFrame((self.relevant_data.iloc[:,0]-self.relevant_data.iloc[:,0].shift(1))/self.relevant_data.iloc[:,0].shift(1))
            for col in self.relevant_data.columns[1:]:
                self.relevant_data_returns[col] = (self.relevant_data[col]-self.relevant_data[col].shift(1))/self.relevant_data[col].shift(1)
            self.relevant_data_returns.fillna(0, inplace=True)
    
    def compute_volatilities(self):
        self.volatilities = pd.DataFrame(self.daily_returns.std(axis=0, skipna =True), columns=["vol_over_the_period"]) # annualiser
        if self.stocks_selection == "covid_resilient":
            for i in range(len(self.relevant_data.index)-1):
                dt1 = self.relevant_data.index[i]
                dt2 = self.relevant_data.index[i+1]
                self.volatilities["vol_"+str(dt1)+"_"+str(dt2)] = self.daily_returns.loc[dt1:dt2, :].std(axis = 0, skipna =True) # a annualise
    def run(self):
        self.load_data()
        self.stocks_selector()
        self.compute_returns()
        self.compute_volatilities()

        a=self.relevant_data_returns.quantile(q=0.25, axis=1)[1]
        rslt = pd.DataFrame(self.relevant_data_returns.loc[["2020-03-18"],:]>a)
        b=self.relevant_data_returns.quantile(q=0.25, axis=1)[2]
        rslt = pd.concat([rslt,self.relevant_data_returns.loc[["2020-10-28"],:]>b])
        c=self.relevant_data_returns.quantile(q=0.25, axis=1)[3]
        rslt = pd.concat([rslt,self.relevant_data_returns.loc[["2021-01-19"],:]>c])

        for i in range(4):
            d = self.volatilities.T.quantile(q=0.25, axis=1)[i]
            rslt = pd.concat([rslt,self.volatilities.T.loc[[self.volatilities.columns[i]],:]<d])
        self.rslt = rslt.astype(int)
        self.rslt.loc["criterion", :] = self.rslt.sum(axis=0)
        self.rslt.sort_values(by="criterion", axis=1, ascending = False, inplace=True)
      